/****************************************************************************
 * Copyright (c) 2010-2012, EBM WebSourcing - All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/
 
package com.ebmwebsourcing.easycommons.xml;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;
import java.io.StringReader;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.stream.FactoryConfigurationError;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.transform.Source;
import javax.xml.transform.TransformerException;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.sax.SAXSource;
import javax.xml.transform.stax.StAXSource;
import javax.xml.transform.stream.StreamSource;

import org.junit.Test;
import org.w3c.dom.Document;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import com.ebmwebsourcing.easycommons.io.IOHelper;

import junit.framework.Assert;

/**
 * 
 * @author Mathieu CARROLLE - EBM WebSourcing
 * 
 */
public class SourceHelperTest {

    private static final Source[] newTestSources(String xmlMessage)
            throws XMLStreamException, FactoryConfigurationError {
        Source streamSource = new StreamSource(new ByteArrayInputStream(xmlMessage.getBytes()));
        Source streamSource2 = new StreamSource(new StringReader(xmlMessage));
        Source domSource = DOMHelper.parseAsDOMSource(new ByteArrayInputStream(xmlMessage
                .getBytes()));
        Source saxSource = new SAXSource(new InputSource(new ByteArrayInputStream(xmlMessage.getBytes())));
        Source staxSource = new StAXSource(
                XMLInputFactory.newInstance().createXMLStreamReader(new ByteArrayInputStream(xmlMessage.getBytes())));
        
        return new Source[] { streamSource, streamSource2, domSource, saxSource, staxSource };
    }
    
    @Test
    public void testSourceToFile() throws Exception {
        for (String xmlMessage : new String[] { "<toto></toto>",
                "<?xml version=\"1.0\" encoding=\"UTF-8\"?><toto></toto>" }) {
            for (Source source : newTestSources(xmlMessage)) {
                File fileToWrite = File.createTempFile("SourceHelperTest", ".xml");
                SourceHelper.toFile(source, fileToWrite);
                
                assertTrue(fileToWrite.exists());
                String actualFileContent = IOHelper.readFileAsString(fileToWrite);
                assertTrue(XMLComparator.isEquivalent(xmlMessage, actualFileContent));
            }
        }
    }
    
    @Test
    public void testSourceToString() throws Exception {
        for (String xmlMessage : new String[] { "<toto></toto>",
                "<?xml version=\"1.0\" encoding=\"UTF-8\"?><toto></toto>" }) {
            for (Source source : newTestSources(xmlMessage)) {
                String sourceAsString = SourceHelper.toString(source);
                assertTrue(XMLComparator.isEquivalent(xmlMessage, sourceAsString));
            }
        }
    }

    @Test(expected = TransformerException.class)
    public void testMalformedStreamSourceToString() throws Exception {
        String xmlMessage = "<tot";
        ByteArrayInputStream bais = new ByteArrayInputStream(xmlMessage.getBytes());
        Source source = new StreamSource(bais);
        SourceHelper.toString(source);
    }

    @Test
    public void testForkAllowedSources() throws Exception {
        for (String xmlMessage : new String[] { "<toto></toto>",
                "<?xml version=\"1.0\" encoding=\"UTF-8\"?><toto></toto>" }) {
            for (Source source : newTestSources(xmlMessage)) {

                if (source instanceof StAXSource)
                    continue;

                Source forkedSource = SourceHelper.fork(source);
                assertEquals(SourceHelper.toString(source), SourceHelper.toString(forkedSource));
            }
        }
    }

    @Test
    public void testForkUnallowedSources() throws Exception {
        for (String xmlMessage : new String[] { "<toto></toto>",
                "<?xml version=\"1.0\" encoding=\"UTF-8\"?><toto></toto>" }) {
            for (Source source : newTestSources(xmlMessage)) {

                if (source instanceof DOMSource)
                    continue;
                if (source instanceof StreamSource)
                    continue;
                if (source instanceof SAXSource)
                    continue;

                try {
                    SourceHelper.fork(source);
                } catch (IllegalArgumentException iae) {
                    continue;
                }
                fail();
            }
        }
    }

    public void testForkMalformedStreamSource() throws TransformerException, IOException {
        String xmlMessage = "<tot";
        ByteArrayInputStream bais = new ByteArrayInputStream(xmlMessage.getBytes());
        Source source = new StreamSource(bais);
        Source forkedSource = SourceHelper.fork(source);
        try {
            SourceHelper.toString(source);
        } catch (TransformerException te1) {
            try {
                SourceHelper.toString(forkedSource);
            } catch (TransformerException te2) {
                return;
            }
        }
        fail();
    }

    
    public void testDOMSourceToInputSource() throws SAXException, IOException, ParserConfigurationException {
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        factory.setNamespaceAware(true);
        Document doc = factory.newDocumentBuilder().parse(Thread.currentThread().getContextClassLoader().getResource("test.xml").openStream());
        DOMSource s = new DOMSource(doc);
        InputSource is = SourceHelper.sourceToInputSource(s, null, s.getSystemId());
        Assert.assertNotNull(is);
    }
    
    public void testSAXSourceToInputSource() throws SAXException, IOException, ParserConfigurationException {
        SAXSource s = new SAXSource(new InputSource(Thread.currentThread().getContextClassLoader().getResource("test.xml").openStream()));
        InputSource is = SourceHelper.sourceToInputSource(s, null, s.getSystemId());
        Assert.assertNotNull(is);
    }
    
    public void testStreamSourceToInputSource() throws SAXException, IOException, ParserConfigurationException {
        StreamSource s = new StreamSource(Thread.currentThread().getContextClassLoader().getResource("test.xml").openStream());
        InputSource is = SourceHelper.sourceToInputSource(s, null, s.getSystemId());
        Assert.assertNotNull(is);
    }

    @Test
    public void testSourceToByteArray() throws Exception {
        for (String xmlMessage : new String[] { "<toto></toto>",
                "<?xml version=\"1.0\" encoding=\"UTF-8\"?><toto></toto>" }) {
            for (Source source : newTestSources(xmlMessage)) {
                byte[] sourceAsString = SourceHelper.toByteArray(source);
                assertTrue(XMLComparator.isEquivalent(xmlMessage, new String(sourceAsString)));
            }
        }
    }

}
